#define std_lang_String_IMPORT
#define std_util_ArraySet_IMPORT

#include <string.h>
#include <stdlib.h>
#include <pthread.h>
#include <std/lang/String.h>
#include <std/util/ArraySet.h>

#define TABLE_SIZE (4095u)

struct Property {
    unsigned int hash;
    size_t length;
};

struct String {
    size_t count;
    struct Property property;
    unsigned char data[0];
};

static union ArraySet table[TABLE_SIZE];

static void
staticInitialize(void)
{
    unsigned int k;

    for (k = 0; k < TABLE_SIZE; ++k) {
	table[k].impl = NULL;
    }
}

__attribute__((constructor)) void
String_staticInitialize(void)
{
    static pthread_once_t once = PTHREAD_ONCE_INIT;

    pthread_once(&once, staticInitialize);
}

static struct Property
getProperty(const char *s)
{
    struct Property p;
    unsigned int h;
    const char *m = s;

    for (h = 0; *m != 0; ++m) {
	h = (h << 5) - h + *(const unsigned char *)m;
    }
    p.hash = h;
    p.length = m - s;
    return p;
}

static struct String *
findSharedString(const char *m, const struct Property *p)
{
    unsigned int i = p->hash % TABLE_SIZE;
    union ArraySet set = table[i];
    size_t k, size;

    if (set.impl == NULL)
	return NULL;
    size = ArraySet_getSize(set);
    for (k = 0; k < size; ++k) {
	struct String *s;

	s = ArraySet_get(set, k);
	if (s->property.hash == p->hash
	    && s->property.length == p->length
	    && memcmp(s->data, m, s->property.length) == 0) {
	    return s;
	}
    }
    return NULL;
}

static int
addTable(struct String *s)
{
    unsigned int i = s->property.hash % TABLE_SIZE;

    if (table[i].impl == NULL && (table[i] = ArraySet_create()).impl == NULL)
	return -1;
    return ArraySet_add(table[i], s);
}

static struct String *
createSharedString(const char *m, const struct Property *p)
{
    struct String *s;
    size_t size;

    size = sizeof(*s) + p->length + 1;
    if ((s = (struct String *)malloc(size)) == NULL)
	return NULL;
    s->count = 1;
    s->property = *p;
    memcpy(s->data, m, p->length + 1);

    if (addTable(s) < 0) {
	free(s);
	return NULL;
    }
    return s;
}

static struct String *
intern(const char *m, const struct Property *p)
{
    struct String *s;

    if ((s = findSharedString(m, p)) != NULL) {
	++(s->count);
	return s;
    }
    return createSharedString(m, p);
}

struct String *
String_intern(const char *m)
{
    struct Property p;

    p = getProperty(m);
    return intern(m, &p);
}

static void
freeSharedString(struct String *s)
{
    unsigned int i = s->property.hash % TABLE_SIZE;

    if (--(s->count) > 0)
	return;
    if (ArraySet_remove(table[i], s) < 0)
	abort();
    free(s);
}

void
String_free(struct String *s)
{
    if (s == NULL)
	return;
    freeSharedString(s);
}

char *
String_getData(struct String *s)
{
    return s->data;
}

size_t
String_getLength(struct String *s)
{
    return s->property.length;
}

void
String_staticEconomize(void)
{
    unsigned int k;

    for (k = 0; k < TABLE_SIZE; ++k) {
	if (table[k].impl == NULL || ArraySet_getSize(table[k]) > 0)
	    continue;
	ArraySet_free(table[k]);
	table[k].impl = NULL;
    }
}
